""":class:`~palladium.interfaces.ModelPersister` implementations.
"""

import gzip
import io
import json
import os
import pickle
import re
from threading import Lock

from sqlalchemy import create_engine
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import LargeBinary
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker

from .interfaces import annotate
from .interfaces import ModelPersister
from .util import logger
from .util import PluggableDecorator
from .util import process_store
from .util import RruleThread
from .util import session_scope


Base = declarative_base()


class File(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that pickles models
    onto the file system, into a given directory.
    """

    def __init__(self, path):
        """
        :param str path:
          The *path* template that I will use to store models,
          e.g. ``/path/to/model-{version}``.
        """
        if path.find('{version}') < 0:
            raise ValueError(
                "Your file persister path must have a {version} placeholder,"
                "e.g., model-{version}.pickle."
                )
        self.path = path

    def read(self, version=None):
        if version is None:
            li = self.list()
            if not li:
                raise IOError("No model available")
            version = li[-1]['version']

        fname = self.path.format(version=version) + '.pkl.gz'
        with gzip.open(fname, 'rb') as f:
            return pickle.load(f)

    def write(self, model):
        last_version = 0
        li = self.list()
        if li:
            last_version = li[-1]['version']

        version = last_version + 1
        li.append(annotate(model, {'version': version}))

        fname = self.path.format(version=version) + '.pkl.gz'
        with gzip.open(fname, 'wb') as f:
            pickle.dump(model, f)

        self._write_md(li)
        return version

    def list(self):
        fname = self.path.format(version='metadata') + '.json'
        if os.path.exists(fname):
            with open(fname, 'r') as f:
                return json.load(f)
        else:
            return []

    def _write_md(self, li):
        fname = self.path.format(version='metadata') + '.json'
        with open(fname, 'w') as f:
            json.dump(li, f, indent=4)


class Database(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that pickles models
    into an SQL database.
    """

    write_lock = Lock()

    class DBModel(Base):
        __tablename__ = 'models'
        version = Column(Integer, primary_key=True)
        metadata_ = Column('metadata', String(length=10 ** 6), nullable=False)
        chunks = relationship(
            'DBModelChunk',
            order_by="DBModelChunk.id",
            )

    class DBModelChunk(Base):
        __tablename__ = 'model_chunks'
        id = Column(Integer, primary_key=True)
        model_version = Column(ForeignKey('models.version'))
        blob = Column(LargeBinary, nullable=False)

    def __init__(self, url, chunk_size=1024 ** 2 * 100):
        """
        :param str url:
          The database *url* that'll be used to make a connection.
          Format follows RFC-1738.  I'll create a table ``models`` to
          store the pickles in if it doesn't exist yet.

        :param int chunk_size:
          The pickled contents of the model are stored inside the
          database in chunks.  The default size is 1024 ** 2 * 100
          (100MB).
        """
        engine = create_engine(url)
        self.engine = engine
        metadata = self.DBModel.metadata
        metadata.bind = engine
        metadata.create_all()
        self.session = scoped_session(sessionmaker(bind=engine))
        self.chunk_size = chunk_size

    def read(self, version=None):
        with session_scope(self.session) as session:
            query = session.query(self.DBModel)
            if not version:
                dbmodel = query.order_by(self.DBModel.version.desc()).first()
            else:
                dbmodel = query.filter_by(version=version).first()

            if dbmodel is not None:
                query2 = session.query(self.DBModelChunk).filter_by(
                    model_version=dbmodel.version
                    ).order_by('id').yield_per(4)
                fileobj = io.BytesIO()
                for chunk in query2:
                    fileobj.write(chunk.blob)
                fileobj.seek(0)
                return pickle.load(gzip.GzipFile(fileobj=fileobj, mode='rb'))

        raise IOError("No model available")

    def write(self, model):
        with self.write_lock:
            return self._write(model)

    def _write(self, model):
        max_version = self.get_max_version()
        if not max_version:
            max_version = 0
        version = max_version + 1

        annotate(model, {'version': version})

        fileobj = io.BytesIO()
        pickle.dump(model, gzip.GzipFile(fileobj=fileobj, mode='wb'))
        data = fileobj.getbuffer()
        chunks = [data[i:i + self.chunk_size]
                  for i in range(0, len(data), self.chunk_size)]

        dbmodel = self.DBModel(
            version=version,
            chunks=[self.DBModelChunk(blob=chunk) for chunk in chunks],
            metadata_=json.dumps(model.__metadata__),
            )

        with session_scope(self.session) as session:
            session.add(dbmodel)

        return version

    def list(self):
        with session_scope(self.session) as session:
            results = session.query(self.DBModel.metadata_).all()
        infos = [json.loads(res[0]) for res in results]
        return sorted(infos, key=lambda x: x['version'])

    def get_max_version(self):
        # We retrieve the max version by hand instead of using an
        # auto-increment because we want to annotate the version
        # number onto the model's metadata.
        with session_scope(self.session) as session:
            query = session.query(self.DBModel.version)
            result = query.order_by(self.DBModel.version.desc()).first()

        if result is not None:
            return result[0]


class CachedUpdatePersister(ModelPersister):
    """A :class:`~palladium.interfaces.ModelPersister` that serves as a
    caching decorator around another `~palladium.interfaces.ModelPersister`
    object.

    Calls to :meth:`~CachedUpdatePersister.read` will look up a model from
    the global ``process_store``, i.e. there is never any actual
    loading involved when calling :meth:`~CachedUpdatePersister.read`.

    To fill the ``process_store`` cache periodically using the return
    value of the underlying :class:`~palladium.interfaces.ModelPersister`'s
    ``read`` method, a dictionary containing keyword arguments to
    :class:`dateutil.rrule.rrule` may be passed.  The cache will then
    be filled periodically according to that recurrence rule.

    If no *update_cache_rrule* is used, the :class:`CachedUpdatePersister`
    will call once and remember the return value of the underlying
    :class:`~palladium.interfaces.ModelPersister`'s ``read`` method during
    initialization.
    """

    cache = process_store
    key = 'model'

    def __init__(self,
                 impl,
                 update_cache_rrule=None
                 ):
        """
        :param ModelPersister impl:
          The underlying (decorated) persister object.

        :param dict update_cache_rrule:
          Optional keyword arguments for a
          :class:`dateutil.rrule.rrule` that determines when the cache
          will be updated.  See :class:`~palladium.util.RruleThread` for
          details.
        """
        self.impl = impl
        self.update_cache_rrule = update_cache_rrule

    def initialize_component(self, config):
        self.use_cache = config.get('__mode__') != 'fit'
        self.thread = None

        if not self.use_cache:
            return

        self.update_cache()
        logger.info("{}: initial fill of cache done.".format(
            self.__class__.__name__))

        if self.update_cache_rrule:
            self.thread = RruleThread(
                func=self.update_cache, rrule=self.update_cache_rrule)
            self.thread.start()

    def read(self, *args, **kwargs):
        if self.use_cache:
            return self.cache[self.key]
        else:
            return self.impl.read(*args, **kwargs)

    @PluggableDecorator('update_model_decorators')
    def update_cache(self, *args, **kwargs):
        model = self.impl.read(*args, **kwargs)
        if model is not None:
            self.cache[self.key] = model
            return model

    @PluggableDecorator('write_model_decorators')
    def write(self, model):
        return self.impl.write(model)

    def list(self):
        return self.impl.list()



"""HTTP API implementation.
"""

import sys
from docopt import docopt
from flask import Flask
from flask import make_response
from flask import request
import numpy as np
import ujson
from werkzeug.exceptions import BadRequest

from . import __version__
from .interfaces import PredictError
from .util import args_from_config
from .util import get_config
from .util import get_metadata
from .util import initialize_config
from .util import logger
from .util import memory_usage_psutil
from .util import PluggableDecorator
from .util import process_store

app = Flask(__name__)


def make_ujson_response(obj, status_code=200):
    """Encodes the given *obj* to json and wraps it in a response.

    :return:
      A Flask response.
    """
    json_encoded = ujson.encode(obj, ensure_ascii=False)
    resp = make_response(json_encoded)
    resp.mimetype = 'application/json'
    resp.content_type = 'application/json; charset=utf-8'
    resp.status_code = status_code
    return resp


class PredictService:
    """A default :class:`palladium.interfaces.PredictService`
    implementation.

    Aims to work out of the box for the most standard use cases.
    Allows overriding of specific parts of its logic by using granular
    methods to compose the work.
    """
    types = {
        'float': float,
        'int': int,
        'str': str,
        'bool': lambda x: x.lower() == 'true',
        }

    def __init__(self, mapping, params=(), predict_proba=False, **kwargs):
        """
        :param mapping:
          A list of query parameters and their type that should be
          included in the request.  These will be processed in the
          :meth:`sample_from_request` method to construct a sample
          that can be used for prediction.  An example that expects
          two request parameters called ``pos`` and ``neg`` that are
          both of type str::

            { ...
              'mapping': [('pos', 'str'), ('neg', 'str')]
            ... }

        :param params:
          Similarly to *mapping*, this is a list of name and type of
          parameters that will be passed to the model's
          :meth:`~palladium.interfaces.Model.predict` method as keyword
          arguments.

        :param predict_proba:
          Instead of returning a single class (the default), when
          *predict_proba* is set to true, the result will instead
          contain a list of class probabilities.
        """
        self.mapping = mapping
        self.params = params
        self.predict_proba = predict_proba
        vars(self).update(kwargs)

    def __call__(self, model, request):
        try:
            return self.do(model, request)
        except Exception as e:
            return self.response_from_exception(e)

    def do(self, model, request):
        """If you need complete control over what the `PredictService`
        would do, you would want to override this method.
        """
        sample, kwargs = self.sample_from_request(model, request)
        y_pred = self.predict(model, sample, **kwargs)
        return self.response_from_prediction(y_pred)

    def sample_from_request(self, model, request):
        values = []
        data = request.args
        for key, type_name in self.mapping:
            value_type = self.types[type_name]
            values.append(value_type(data[key]))
        params = {}
        for key, type_name in self.params:
            value_type = self.types[type_name]
            if key in data:
                params[key] = value_type(data[key])
            elif hasattr(model, key):
                params[key] = getattr(model, key)
        return np.array([values], dtype=object), params

    def predict(self, model, sample, **kwargs):
        if self.predict_proba:
            return model.predict_proba(sample, **kwargs)
        else:
            return model.predict(sample, **kwargs)

    def response_from_prediction(self, y_pred):
        response = {
            'metadata': get_metadata(),
            'result': y_pred.tolist()[0],
            }
        return make_ujson_response(response, status_code=200)

    def response_from_exception(self, exc):
        if isinstance(exc, PredictError):
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=exc.error_code,
                    error_message=exc.error_message,
                    status="ERROR"
                )
            }, status_code=500)
        elif isinstance(exc, BadRequest):
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=-1,
                    error_message="BadRequest: {}".format(exc.args),
                    status="ERROR"
                )
            }, status_code=400)
        else:
            logger.exception("Unexpected error")
            return make_ujson_response({
                'metadata': get_metadata(
                    error_code=-1,
                    error_message="{}: {}".format(
                        exc.__class__.__name__, str(exc)),
                    status="ERROR"
                )
            }, status_code=500)


@app.route('/predict')
@PluggableDecorator('predict_decorators')
@args_from_config
def predict(model_persister, predict_service):
    try:
        model = model_persister.read()
        response = predict_service(model, request)
    except Exception as exc:
        logger.exception("Unexpected error")
        response = make_ujson_response({
            "status": "ERROR",
            "error_code": -1,
            "error_message": "{}: {}".format(exc.__class__.__name__, str(exc)),
        }, status_code=500)

    return response


@app.route('/alive')
@PluggableDecorator('alive_decorators')
@args_from_config
def alive(alive=None):
    if alive is None:
        alive = {}

    info = {
        'memory_usage': int(memory_usage_psutil()),
        'palladium_version': __version__,
        }

    info['service_metadata'] = get_config().get('service_metadata', {})

    status_code = 200
    for attr in alive.get('process_store_required', ()):
        obj = process_store.get(attr)
        if obj is not None:
            obj_info = {}
            obj_info['updated'] = process_store.mtime['model'].isoformat()
            if hasattr(obj, '__metadata__'):
                obj_info['metadata'] = obj.__metadata__
            info[attr] = obj_info
        else:
            info[attr] = "N/A"
            status_code = 503

    return make_ujson_response(info, status_code=status_code)


def devserver_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
Run a server for development.

Usage:
  palladium-devserver [options]

Options:
  -h --help                  Show this screen.

  --host=<host>           The host to use [default: 0.0.0.0].

  --port=<port>           The port to use [default: 5000].

  --debug=<debug>         Whether or not to use debug mode [default: 0].
"""
    arguments = docopt(__doc__, argv=argv)
    initialize_config()
    app.run(
        host=arguments['--host'],
        port=int(arguments['--port']),
        debug=int(arguments['--debug']),
        )



import os
from threading import Thread
from unittest.mock import MagicMock
from unittest.mock import patch

import numpy as np
from pandas import DataFrame
import pytest

dummy_dataframe = DataFrame({
    'datacol1': [10, 11, 12, 13, 14],
    'datacol2': [20.0, 21.0, 22.0, 23.0, 24.0],
    'targetcol': [0, 1, 2, 3, 4],
    })


class TestTable:
    @pytest.fixture
    def Table(self):
        from palladium.dataset import Table
        return Table

    def test_it(self, Table):
        with patch("palladium.dataset.Table.pandas_read") as read_table:
            read_table.return_value = dummy_dataframe[3:5]  # simulate skiprows
            dataset = Table('mypath', 'targetcol', some='keyword', skiprows=3)
            data, target = dataset()

        read_table.assert_called_with('mypath', some='keyword', skiprows=3)
        assert len(data) == len(target) == 2
        assert data.tolist() == [[13, 23.0], [14, 24.0]]
        assert target.tolist() == [3, 4]

    def test_ndarray_false(self, Table):
        with patch("palladium.dataset.Table.pandas_read") as read_table:
            read_table.return_value = dummy_dataframe[3:5]
            dataset = Table('mypath', 'targetcol', some='keyword',
                            skiprows=3, ndarray=False)  # simulate skiprows
            data, target = dataset()

        assert data['datacol1'].tolist() == [13, 14]
        assert data['datacol2'].tolist() == [23.0, 24.0]
        assert target.tolist() == [3, 4]

    def test_no_slice(self, Table):
        with patch("palladium.dataset.Table.pandas_read") as read_table:
            read_table.return_value = dummy_dataframe
            dataset = Table('mypath', 'targetcol', some='keyword')
            data, target = dataset()

        read_table.assert_called_with('mypath', some='keyword')
        assert len(data) == len(target) == len(dummy_dataframe)
        assert data.tolist() == [
            [10, 20.0], [11, 21.0], [12, 22.0], [13, 23.0], [14, 24.0],
            ]
        assert target.tolist() == [0, 1, 2, 3, 4]

    def test_table_no_target(self, Table):
        with patch("palladium.dataset.Table.pandas_read") as read_table:
            read_table.return_value = dummy_dataframe
            dataset = Table('mypath', some='keyword')
            data, target = dataset()

        read_table.assert_called_with('mypath', some='keyword')
        assert len(data) == len(dummy_dataframe)
        assert target is None


class TestSQL:
    @pytest.fixture
    def SQL(self):
        from palladium.dataset import SQL
        return SQL

    @pytest.fixture
    def sql(self, request, SQL):
        path = '/tmp/palladium.testing-{}.sqlite'.format(os.getpid())
        request.addfinalizer(lambda: os.remove(path))
        sql = SQL(
            url='sqlite:///{}'.format(path),
            sql='select age, weight, salary from employee',
            target_column='salary',
            )

        connection = sql.engine.connect()
        connection.execute("""
        CREATE TABLE EMPLOYEE (
           id INT PRIMARY KEY,
           name TEXT,
           age INT,
           weight REAL,
           salary REAL
           )
           """)
        for values in [
            "(1, 'James', 24, 60.0, 10000.0)",
            "(2, 'Guido', 33, 73.0, 20000.0)",
            "(3, 'Handsome Jack', 27, 67.5, 35000.0)",
            ]:
            connection.execute(
                "INSERT INTO EMPLOYEE VALUES {}".format(values))

        return sql

    def test_it(self, sql):
        X, y = sql()
        assert X.tolist() == [
            [24., 60.],
            [33., 73.],
            [27., 67.5],
            ]
        assert y.tolist() == [
            10000.0,
            20000.0,
            35000.0,
            ]

    def test_ndarray_false(self, sql):
        sql.ndarray = False
        X, y = sql()
        assert X.values.tolist() == [
            [24., 60.],
            [33., 73.],
            [27., 67.5],
            ]
        assert y.values.tolist() == [
            10000.0,
            20000.0,
            35000.0,
            ]

    def test_concurrency(self, sql):
        threads = [Thread(target=sql) for i in range(5)]
        [th.start() for th in threads]
        [th.join() for th in threads]


def test_empty_dataset_loader():
    from palladium.dataset import EmptyDatasetLoader
    edl = EmptyDatasetLoader()
    X, y = edl()
    assert X is None
    assert y is None


class TestScheduledDatasetLoader:
    @pytest.fixture
    def ScheduledDatasetLoader(self, process_store):
        from palladium.dataset import ScheduledDatasetLoader
        return ScheduledDatasetLoader

    @pytest.fixture
    def loader(self, ScheduledDatasetLoader, config):
        loader = ScheduledDatasetLoader(
            MagicMock(),
            {
                'freq': 'DAILY',
                'dtstart': '2014-10-30T13:21:18',
                },
            )
        loader.initialize_component(config)
        return loader

    def test_call(self, process_store, loader):
        assert loader() is loader.impl.return_value

    def test_call_custom_value(self, process_store, loader):
        process_store['data'] = ('X', 'y')
        assert loader() == ('X', 'y')

    def test_update_cache(self, loader):
        loader.impl.return_value = ('bla', 'bla')
        assert loader.update_cache() == ('bla', 'bla')
        assert loader() == ('bla', 'bla')
        assert loader.impl.call_count == 2



from datetime import datetime
import json
from unittest.mock import Mock
from unittest.mock import patch

import dateutil.parser
from flask import request
import numpy as np
import pytest
from werkzeug.exceptions import BadRequest


class TestPredictService:
    @pytest.fixture
    def PredictService(self):
        from palladium.server import PredictService
        return PredictService

    def test_functional(self, PredictService, flask_app):
        model = Mock()
        model.threshold = 0.3
        model.size = 10
        # needed as hasattr would evaluate to True otherwise
        del model.threshold2
        del model.size2
        model.predict.return_value = np.array(['class1'])
        service = PredictService(
            mapping=[
                ('sepal length', 'float'),
                ('sepal width', 'float'),
                ('petal length', 'float'),
                ('petal width', 'float'),
                ('color', 'str'),
                ('age', 'int'),
                ('active', 'bool'),
                ('austrian', 'bool'),
                ],
            params=[
                ('threshold', 'float'),  # default will be overwritten
                ('size', 'int'),  # not provided, default value kept
                ('threshold2', 'float'),  # will be used, no default value
                ('size2', 'int'),  # not provided, no default value
                ])

        with flask_app.test_request_context():
            with patch('palladium.util.get_config') as get_config:
                meta_dict = {
                    'service_metadata': {
                        'service_name': 'iris',
                        'service_version': '0.1'
                    }
                }
                get_config.return_value = meta_dict
                resp = service(
                    model,
                    Mock(args=dict([
                        ('sepal length', '5.2'),
                        ('sepal width', '3.5'),
                        ('petal length', '1.5'),
                        ('petal width', '0.2'),
                        ('color', 'purple'),
                        ('age', '1'),
                        ('active', 'True'),
                        ('austrian', 'False'),
                        ('threshold', '0.7'),
                        ('threshold2', '0.8'),
                    ]))
                )

        assert (model.predict.call_args[0][0] ==
                np.array([[5.2, 3.5, 1.5, 0.2,
                           'purple', 1, True, False]], dtype='object')).all()
        assert model.predict.call_args[1]['threshold'] == 0.7
        assert model.predict.call_args[1]['size'] == 10
        assert model.predict.call_args[1]['threshold2'] == 0.8
        assert 'size2' not in model.predict.call_args[1]
        assert resp.status_code == 200
        expected_resp_data = {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                "service_name": "iris",
                "service_version": "0.1",
                },
            "result": "class1"
            }

        assert json.loads(resp.get_data(as_text=True)) == expected_resp_data

    def test_bad_request(self, PredictService, flask_app):
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                with patch('palladium.util.get_config') as get_config:
                    meta_dict = {
                        'service_metadata': {}
                    }
                    get_config.return_value = meta_dict
                    bad_request = BadRequest()
                    bad_request.args = ('daniel',)
                    psd.side_effect = bad_request
                    resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 400
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": -1,
                "error_message": "BadRequest: ('daniel',)"
                }
            }

    def test_predict_error(self, PredictService, flask_app):
        from palladium.interfaces import PredictError
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                with patch('palladium.util.get_config') as get_config:
                    meta_dict = {
                        'service_metadata': {}
                    }
                    get_config.return_value = meta_dict
                    psd.side_effect = PredictError("mymessage", 123)
                    resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": 123,
                "error_message": "mymessage",
                }
            }

    def test_generic_error(self, PredictService, flask_app):
        predict_service = PredictService(mapping=[])
        model = Mock()
        request = Mock()
        with patch.object(predict_service, 'do') as psd:
            with flask_app.test_request_context():
                with patch('palladium.util.get_config') as get_config:
                    meta_dict = {
                        'service_metadata': {}
                    }
                    get_config.return_value = meta_dict
                    psd.side_effect = KeyError("model")
                    resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "metadata": {
                "status": "ERROR",
                "error_code": -1,
                "error_message": "KeyError: 'model'",
                }
            }

    def test_sample_from_request(self, PredictService):
        predict_service = PredictService(
            mapping=[
                ('name', 'str'),
                ('sepal width', 'int'),
                ],
            )

        model = Mock()
        request = Mock(args={'name': 'myflower', 'sepal width': 3})
        sample, params = predict_service.sample_from_request(model, request)
        assert sample[0][0] == 'myflower'
        assert sample[0][1] == 3

    def test_probas(self, PredictService, flask_app):
        model = Mock()
        model.predict_proba.return_value = np.array([[0.1, 0.5, 0.4]])
        predict_service = PredictService(mapping=[], predict_proba=True)
        with flask_app.test_request_context():
            with patch('palladium.util.get_config') as get_config:
                meta_dict = {
                    'service_metadata': {}
                }
                get_config.return_value = meta_dict
                resp = predict_service(model, request)
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 200
        assert resp_data == {
            "metadata": {
                "status": "OK",
                "error_code": 0,
                },
            "result": [0.1, 0.5, 0.4],
            }


class TestPredict:
    @pytest.fixture
    def predict(self):
        from palladium.server import predict
        return predict

    def test_predict_functional(self, config, flask_app, flask_client):
        from palladium.server import make_ujson_response
        model_persister = config['model_persister'] = Mock()
        predict_service = config['predict_service'] = Mock()
        with flask_app.test_request_context():
            predict_service.return_value = make_ujson_response(
                'a', status_code=200)

        model = model_persister.read()

        resp = flask_client.get(
            'predict?sepal length=1.0&sepal width=1.1&'
            'petal length=0.777&petal width=5')

        resp_data = json.loads(resp.get_data(as_text=True))

        assert resp_data == 'a'
        assert resp.status_code == 200
        with flask_app.test_request_context():
            predict_service.assert_called_with(model, request)

    def test_unknown_exception(self, predict, flask_app):
        model_persister = Mock()
        model_persister.read.side_effect = KeyError('model')

        with flask_app.test_request_context():
            resp = predict(model_persister, Mock())
        resp_data = json.loads(resp.get_data(as_text=True))
        assert resp.status_code == 500
        assert resp_data == {
            "status": "ERROR",
            "error_code": -1,
            "error_message": "KeyError: 'model'",
            }


class TestAliveFunctional:
    def test_empty_process_state(self, config, flask_client):
        config['service_metadata'] = {'hello': 'world'}
        resp = flask_client.get('alive')
        assert resp.status_code == 200
        resp_data = json.loads(resp.get_data(as_text=True))

        assert sorted(resp_data.keys()) == ['memory_usage',
                                            'palladium_version',
                                            'service_metadata']
        assert resp_data['service_metadata'] == config['service_metadata']

    def test_filled_process_state(self, config, process_store, flask_client):
        config['alive'] = {'process_store_required': ('model', 'data')}

        before = datetime.now()
        process_store['model'] = Mock(__metadata__={'hello': 'is it me'})
        process_store['data'] = Mock(__metadata__={'bye': 'not you'})
        after = datetime.now()

        resp = flask_client.get('alive')
        assert resp.status_code == 200
        resp_data = json.loads(resp.get_data(as_text=True))

        model_updated = dateutil.parser.parse(resp_data['model']['updated'])
        data_updated = dateutil.parser.parse(resp_data['data']['updated'])
        assert before < model_updated < after
        assert resp_data['model']['metadata'] == {'hello': 'is it me'}
        assert before < data_updated < after
        assert resp_data['data']['metadata'] == {'bye': 'not you'}

    def test_missing_process_state(self, config, process_store, flask_client):
        config['alive'] = {'process_store_required': ('model', 'data')}
        process_store['model'] = Mock(__metadata__={'hello': 'is it me'})

        resp = flask_client.get('alive')
        assert resp.status_code == 503
        resp_data = json.loads(resp.get_data(as_text=True))

        assert resp_data['model']['metadata'] == {'hello': 'is it me'}
        assert resp_data['data'] == 'N/A'



import pytest


@pytest.fixture
def flask_client(flask_app):
    return flask_app.test_client()


@pytest.fixture
def process_store(request):
    from palladium.util import process_store

    orig = process_store.copy()
    process_store.clear()
    request.addfinalizer(
        lambda: (process_store.clear(), process_store.update(orig)))
    return process_store



from datetime import datetime
from time import sleep
from unittest.mock import Mock
from unittest.mock import mock_open
from unittest.mock import patch

from dateutil import rrule
import pytest


class MyDummyComponent:
    def __init__(self, arg1, arg2='blargh', subcomponent=None):
        self.arg1 = arg1
        self.arg2 = arg2
        self.subcomponent = subcomponent
        self.initialize_component_arg = None

    def initialize_component(self, config):
        self.initialize_component_arg = config


def test_create_component():
    from palladium.util import create_component

    result = create_component({
        '__factory__': 'palladium.tests.test_util.MyDummyComponent',
        'arg1': 3,
        })
    assert isinstance(result, MyDummyComponent)
    assert result.arg1 == 3
    assert result.arg2 == 'blargh'


def test_config_class_keyerror():
    from palladium.util import Config
    with pytest.raises(KeyError) as e:
        Config({})['invalid']
    assert "Maybe you forgot to set" in str(e.value)


class TestResolveDottedName:
    @pytest.fixture
    def resolve_dotted_name(self):
        from palladium.util import resolve_dotted_name
        return resolve_dotted_name

    def test_with_colon(self, resolve_dotted_name):
        dotted = 'palladium.tests.test_util:TestResolveDottedName.test_with_colon'
        assert (resolve_dotted_name(dotted) is
                TestResolveDottedName.test_with_colon)

    def test_with_dots(self, resolve_dotted_name):
        dotted = 'palladium.tests.test_util.TestResolveDottedName'
        assert (resolve_dotted_name(dotted) is
                TestResolveDottedName)


class TestInitializeConfigImpl:
    @pytest.fixture
    def _initialize_config(self):
        from palladium.util import _initialize_config
        return _initialize_config

    def test_initialize_config(self, _initialize_config):
        config = {
            'mycomponent': {
                '__factory__': 'palladium.tests.test_util.MyDummyComponent',
                'arg1': 3,
                'arg2': {'no': 'factory'},
                'subcomponent': {
                    '__factory__': 'palladium.tests.test_util.MyDummyComponent',
                    'arg1': {
                        'subsubcomponent': {
                            '__factory__':
                            'palladium.tests.test_util.MyDummyComponent',
                            'arg1': 'wobwob',
                            'arg2': 9,
                            },
                        },
                    'arg2': 6,
                    },
                },
            'mylistofcomponents': [{
                '__factory__': 'palladium.tests.test_util.MyDummyComponent',
                'arg1': 'wobwob',
                },
                'somethingelse',
                ],
            'myconstant': 42,
            }

        config = _initialize_config(config)
        assert config['myconstant'] == 42

        mycomponent = config['mycomponent']
        assert isinstance(mycomponent, MyDummyComponent)
        assert mycomponent.arg1 == 3
        assert mycomponent.arg2 == {'no': 'factory'}
        assert mycomponent.initialize_component_arg is config

        subcomponent = mycomponent.subcomponent
        assert isinstance(subcomponent, MyDummyComponent)
        assert subcomponent.arg2 == 6
        assert subcomponent.initialize_component_arg is config

        subsubcomponent = subcomponent.arg1['subsubcomponent']
        assert isinstance(subsubcomponent, MyDummyComponent)
        assert subsubcomponent.arg1 == 'wobwob'
        assert subsubcomponent.arg2 == 9
        assert subsubcomponent.initialize_component_arg is config

        mylistofcomponents = config['mylistofcomponents']
        assert len(mylistofcomponents) == 2
        assert isinstance(mylistofcomponents[0], MyDummyComponent)
        assert mylistofcomponents[0].arg1 == 'wobwob'
        assert mylistofcomponents[1] == 'somethingelse'

    def test_initialize_config_logging(self, _initialize_config):
        with patch('palladium.util.dictConfig') as dictConfig:
            _initialize_config({'logging': 'yes, please'})
            dictConfig.assert_called_with('yes, please')


class TestInitializeConfig:
    @pytest.fixture
    def initialize_config(self):
        from palladium.util import initialize_config
        return initialize_config

    def test_initialize_config_extra(self, config,
                                     initialize_config):
        config.clear()
        initialize_config(two='three')
        assert config['two'] == 'three'

    def test_initialize_config_already_initialized(self, config,
                                                   initialize_config):
        config.clear()
        config.initialized = True
        with pytest.raises(RuntimeError):
            initialize_config(two='three')


class TestGetConfig:
    @pytest.fixture
    def get_config(self):
        from palladium.util import get_config
        return get_config

    def test_read(self, get_config, config):
        config.initialized = False
        config_in = {
            'mycomponent': {
                '__factory__': 'palladium.tests.test_util.MyDummyComponent',
                'arg1': 3,
                },
            'myconstant': 42,
            }

        with patch('palladium.util.open',
                   mock_open(read_data=str(config_in)),
                   create=True):
            with patch('palladium.util.os.environ', {'PALLADIUM_CONFIG': 'somepath'}):
                config_new = get_config()

            assert config_new['myconstant'] == config_in['myconstant']
            mycomponent = config_new['mycomponent']
            assert isinstance(mycomponent, MyDummyComponent)
            assert mycomponent.arg1 == 3

    def test_read_environ(self, get_config, config):
        config.initialized = False
        config_in_str = """
{
'mycomponent': {
    '__factory__': 'palladium.tests.test_util.MyDummyComponent',
    'arg1': 3,
    'arg2': "{}:{}".format(environ['PALLADIUM_DB_IP'],
                           environ['PALLADIUM_DB_PORT']),
    },
}"""

        with patch('palladium.util.open',
                   mock_open(read_data=config_in_str),
                   create=True):
            with patch('palladium.util.os.environ', {
                    'PALLADIUM_CONFIG': 'somepath',
                    'PALLADIUM_DB_IP': '192.168.0.1',
                    'PALLADIUM_DB_PORT': '666',
            }):
                config_new = get_config()

            mycomponent = config_new['mycomponent']
            assert isinstance(mycomponent, MyDummyComponent)
            assert mycomponent.arg1 == 3
            assert mycomponent.arg2 == '192.168.0.1:666'


def test_args_from_config(config):
    from palladium.util import args_from_config

    args = []
    config.update({
        'arg1': 'config1',
        'arg2': 'config2',
        'arg4': 'config4',
        })

    @args_from_config
    def myfunc(arg1, arg2, arg3, arg4='default4', arg5='default5'):
        args.append((arg1, arg2, arg3, arg4, arg5))

    myfunc(arg3='myarg3')
    assert args[-1] == ('config1', 'config2', 'myarg3', 'config4', 'default5')

    myfunc(arg2='myarg2', arg3='myarg3', arg5='myarg5')
    assert args[-1] == ('config1', 'myarg2', 'myarg3', 'config4', 'myarg5')

    myfunc('myarg1', arg3='myarg3')
    assert args[-1] == ('myarg1', 'config2', 'myarg3', 'config4', 'default5')

    with pytest.raises(TypeError) as e:
        myfunc()
    assert "Maybe you forgot to set" in str(e.value)


class TestProcessStore:
    @pytest.fixture
    def store(self):
        from palladium.util import ProcessStore
        return ProcessStore()

    def test_get(self, store):
        assert store.get('somekey') is None
        store['somekey'] = '1'
        assert store.get('somekey') == '1'

    def test_getitem(self, store):
        with pytest.raises(KeyError):
            store['somekey']
        store['somekey'] = '1'
        assert store['somekey'] == '1'

    def test_in(self, store):
        assert not 'somekey' in store
        store['somekey'] = '1'
        assert 'somekey' in store

    def test_setitem(self, store):
        store['somekey'] = '1'
        assert store['somekey'] == '1'
        store['somekey'] = '2'
        assert store['somekey'] == '2'

    def test_delitem(self, store):
        store['somekey'] = '1'
        del store['somekey']
        with pytest.raises(KeyError):
            store['somekey']

    def test_mtime_no_entry(self, store):
        with pytest.raises(KeyError):
            assert store.mtime['somekey']

    def test_mtime_setitem(self, store):
        dt0 = datetime.now()
        store['somekey'] = '1'
        dt1 = datetime.now()
        assert dt0 < store.mtime['somekey'] < dt1
        store['somekey'] = '2'
        dt2 = datetime.now()
        assert dt1 < store.mtime['somekey'] < dt2

    def test_mtime_delitem(self, store):
        store['somekey'] = '1'
        del store['somekey']
        with pytest.raises(KeyError):
            store.mtime['somekey']

    def test_init(self, store):
        store = store.__class__({'somekey': '1'})
        assert store['somekey'] == '1'
        assert 'somekey' in store.mtime

    def test_clear(self, store):
        store['somekey'] = '1'
        store.clear()
        assert 'somekey' not in store
        assert 'somekey' not in store.mtime


class TestRruleThread:
    @pytest.fixture
    def RruleThread(self):
        from palladium.util import RruleThread
        return RruleThread

    def test_rrule_from_dict(self, RruleThread):
        func = Mock()
        now = datetime.now()
        rrule_info = {
            'freq': 'DAILY',
            'dtstart': '2014-10-30T13:21:18',
            }
        expected = rrule.rrule(
            rrule.DAILY, dtstart=datetime(2014, 10, 30, 13, 21, 18))

        thread = RruleThread(func, rrule_info)
        assert thread.rrule.after(now) == expected.after(now)

    def test_last_execution(self, RruleThread):
        func = Mock()
        thread = RruleThread(func, rrule.rrule(
            rrule.MONTHLY,
            bymonthday=1,
            dtstart=datetime(2014, 10, 30, 13, 21, 18)))
        thread.last_execution = datetime(2014, 10, 30, 13, 21, 18)
        thread.start()
        sleep(0.005)
        assert func.call_count == 1

    def test_func_raises(self, RruleThread):
        func = Mock(__name__='buggy')
        func.side_effect = KeyError
        thread = RruleThread(func, rrule.rrule(
            rrule.MONTHLY,
            bymonthday=1,
            dtstart=datetime(2014, 10, 30, 13, 21, 18)))
        thread.last_execution = datetime(2014, 10, 30, 13, 21, 18)

        with patch('palladium.util.logger') as logger:
            thread.start()
            sleep(0.005)
            assert func.call_count == 1
            assert logger.exception.call_count == 1

    def test_sleep_between_checks(self, RruleThread):
        func = Mock()
        rr = Mock()
        rr.between.return_value = False
        thread = RruleThread(func, rr, sleep_between_checks=0.0010)
        thread.start()
        sleep(0.005)
        assert func.call_count == 0
        assert rr.between.call_count > 1


def dec1(func):
    def inner(a, b):
        """dec1"""
        return func(a, b) * 10
    return inner


def dec2(func):
    def inner(a, b):
        """dec2"""
        return func(a, b) + 2
    return inner


class TestPluggableDecorator:
    @pytest.fixture
    def PluggableDecorator(self):
        from palladium.util import PluggableDecorator
        return PluggableDecorator

    def test_no_config(self, PluggableDecorator, config):
        decorator = PluggableDecorator('decorator_list')

        @decorator
        def myfunc(a, b):
            """myfunc"""
            return a + b

        result = myfunc(2, 3)
        assert result == 5
        assert myfunc.__doc__ == "myfunc"

    def test_two_decorators(self, PluggableDecorator, config):
        config['decorator_list'] = [
            'palladium.tests.test_util.dec1',
            'palladium.tests.test_util.dec2']
        decorator = PluggableDecorator('decorator_list')

        @decorator
        def myfunc(a, b):
            """myfunc"""
            return a + b

        result = myfunc(2, 3)
        assert result == 52
        assert myfunc.__doc__ == "myfunc"

    def test_wrapped_empty_list(self, PluggableDecorator, config):
        def myfunc(a, b):
            """myfunc"""
            return a + b

        myfunc2 = PluggableDecorator('decorator_list')(myfunc)
        assert hasattr(myfunc, '__wrapped__') is False
        assert myfunc2.__wrapped__ is myfunc

    def test_wrapped(self, PluggableDecorator, config):
        config['decorator_list'] = [
            'palladium.tests.test_util.dec1',
            'palladium.tests.test_util.dec2']

        def myfunc(a, b):
            """myfunc"""
            return a + b

        myfunc2 = PluggableDecorator('decorator_list')(myfunc)
        assert hasattr(myfunc, '__wrapped__') is False
        assert hasattr(myfunc2, '__wrapped__') is True
        assert myfunc2.__wrapped__ is myfunc


class TestSessionScope:

    def test_success(self):
        from palladium.util import session_scope
        session = Mock()
        with session_scope(session):
            session.query()
        assert session.commit.call_count == 1
        assert session.rollback.call_count == 0
        assert session.close.call_count == 1

    def test_exception(self):
        from palladium.util import session_scope
        session = Mock()
        with pytest.raises(KeyError):
            with session_scope(session):
                raise KeyError('Error')
        assert session.commit.call_count == 0
        assert session.rollback.call_count == 1
        assert session.close.call_count == 1


class TestApplyKwargs:
    @pytest.fixture
    def apply_kwargs(self):
        from palladium.util import apply_kwargs
        return apply_kwargs

    @pytest.fixture
    def myfunc(self):
        calls = []

        def myfunc(one, two, three='three'):
            calls.append((one, two, three))

        myfunc.calls = calls
        return myfunc

    def test_it(self, apply_kwargs, myfunc):
        apply_kwargs(myfunc, one=1, two='two', four=4)
        assert myfunc.calls == [(1, 'two', 'three')]

    def test_typeerror(self, apply_kwargs, myfunc):
        with pytest.raises(TypeError):
            apply_kwargs(myfunc, two='two', four=4)


class TestPartial:
    calls = []

    @classmethod
    def myfunc(cls, one, two, three='three'):
        cls.calls.append((one, two, three))

    @pytest.fixture
    def Partial(self):
        from palladium.util import Partial
        return Partial

    def test_it(self, Partial):
        Partial(self.myfunc, two=2)(one='one')
        assert self.calls == [('one', 2, 'three')]
        self.__class__.calls = []

    def test_dotted(self, Partial):
        Partial('palladium.tests.test_util:TestPartial.myfunc', two=2)(one='one')
        assert self.calls == [('one', 2, 'three')]
        self.__class__.calls = []



from random import random

import pandas
import pytest


class TestDiskCache:
    @pytest.fixture
    def diskcache(self):
        from palladium.cache import diskcache
        return diskcache

    def test_it(self, diskcache):
        called = []

        @diskcache(lambda x: 'key_{}'.format(x))
        def squareit(x):
            called.append(x)
            return pandas.DataFrame([x ** 2])

        cache = squareit.__cache__
        cache.filename_tmpl = '/tmp/{module}.{func}-cache-{key}-%s.cache' % (
            str(random()),)

        assert (squareit(2).squeeze(),
                squareit(3).squeeze(),
                squareit(3).squeeze()) == (4, 9, 9)
        assert called == [2, 3]

    def test_it_no_compute_key(self, diskcache):
        called = []

        @diskcache()
        def squareit(x):
            called.append(x)
            return pandas.DataFrame([x ** 2])

        cache = squareit.__cache__
        cache.filename_tmpl = '/tmp/{module}.{func}-cache-{key}-%s.cache' % (
            str(random()),)

        assert (squareit(2).squeeze(),
                squareit(3).squeeze(),
                squareit(3).squeeze()) == (4, 9, 9)
        assert called == [2, 3]

    def test_it_ignore(self, diskcache):
        called = []

        @diskcache(ignore=True)
        def squareit(x):
            called.append(x)
            return pandas.DataFrame([x ** 2])

        cache = squareit.__cache__
        cache.filename_tmpl = '/tmp/{module}.{func}-cache-{key}-%s.cache' % (
            str(random()),)

        assert (squareit(2).squeeze(),
                squareit(3).squeeze(),
                squareit(3).squeeze()) == (4, 9, 9)
        assert called == [2, 3, 3]


class TestPickleDiskCache(TestDiskCache):
    @pytest.fixture
    def diskcache(self):
        from palladium.cache import picklediskcache
        return picklediskcache



"""Interfaces defining the behaviour of Palladium's components.
"""

from abc import abstractmethod
from abc import ABCMeta

from sklearn.base import BaseEstimator


def annotate(obj, metadata=None):
    base_metadata = getattr(obj, '__metadata__', {})
    if metadata is not None:
        base_metadata.update(metadata)
        obj.__metadata__ = base_metadata
    return obj.__metadata__


class DatasetLoader(metaclass=ABCMeta):
    """A :class:`~palladium.interfaces.DatasetLoader` is responsible for
    loading datasets for use in training and evaluation.
    """

    @abstractmethod
    def __call__(self):
        """Loads the data and returns a tuple *(data, target)*, or
        *(X, y)*.

        :return:
          A tuple *(data, target*).

          *data* is a two dimensional numpy array with shape n x m
          (one row per example).

          *target* is a one dimensional array with n target values.

          *target* may be ``None`` if there is no target value,
          e.g. in an unsupervised learning setting.

        :rtype: tuple
        """


class CrossValidationGenerator(metaclass=ABCMeta):
    """A :class:`CrossValidationGenerator` provides train/test indices
    to split data in train and validation sets.

    :class:`CrossValidationGenerator` corresponds to the *cross
    validation generator* interface of scikit-learn.
    """

    @abstractmethod
    def __iter__(self):
        """
        :return:
          Tuples of train/test indices.
        """


class Model(BaseEstimator, metaclass=ABCMeta):
    """A :class:`Model` can be :meth:`~Model.fit` to data and can be
    used to :meth:`~Model.predict` data.

    :class:`Model` corresponds to the *estimators* interface of
    scikit-learn.
    """

    @abstractmethod
    def fit(self, X, y=None):
        """Fit to data array *X* and possibly a target array *y*.

        :return: self
        """

    @abstractmethod
    def predict(self, X, **kw):
        """Predict classes for data array *X* with shape n x m.

        Some models may accept additional keyword arguments.

        :return:
          A numpy array of length n with the predicted classes (for
          classification problems) or numeric values (for regression
          problems).

        :raises:
          May raise a :class:`PredictError` to indicate that some
          condition made it impossible to deliver a prediction.
        """

    def predict_proba(self, X, **kw):
        """Predict probabilities for data array *X* with shape n x m.

        :return:
          A numpy array of length n x c with a list class
          probabilities per sample.

        :raises:
          :class:`NotImplementedError` if not applicable.
        """
        raise NotImplementedError()  # pragma: no cover


class PredictError(Exception):
    """Raised by :meth:`Model.predict` to indicate that some condition
    made it impossible to deliver a prediction.
    """
    def __init__(self, error_message, error_code=-1):
        self.error_message = error_message
        self.error_code = error_code

    def __str__(self):
        return "{} ({})".format(self.error_message, self.error_code)


class ModelPersister(metaclass=ABCMeta):
    @abstractmethod
    def read(self, version=None):
        """Returns a :class:`Model` instance.

        :param str version:
          *version* may be used to read a specific version of a model.
          If *version* is ``None``, returns the latest model.

        :return:
          The model object.

        :raises:
          IOError if no model was available.
        """

    @abstractmethod
    def write(self, model):
        """Persists a :class:`Model` and returns a new version number.

        It is the :class:`ModelPersister`'s responsibility to annotate
        the 'version' information onto the model before it is saved.
        """

    @abstractmethod
    def list(self):
        """List metadata of all available models.

        :return:
          A list of dicts, with each dict containing information about
          one of the available models.  Each dict is guaranteed to
          contain the ``version`` key, which is the same version
          number that :meth:`ModelPerister.read` accepts for loading
          specific models.
        """


class PredictService(metaclass=ABCMeta):
    """Responsible for producing the output for the '/predict' HTTP
    endpoint.
    """
    @abstractmethod
    def __call__(self, model, request):
        """
        Use the model to run a prediction with the requested data.

        :param model:
          The :class:`~Model` instance to use for making predictions.
        :param request:
          A werkzeug ``request`` object.  A dictionary with query
          parameters is available at *request.args*.

        :return:
          A werkzeug ``response`` object.  It is the
          :class:`PredictService`'s responsiblity to return
          appropriate status codes and data in case of error.
        """



import numpy
import os
from pandas import DataFrame
from pandas import Series

from unittest.mock import MagicMock
from unittest.mock import Mock

import pytest


pytest.importorskip("rpy2")
from pandas.rpy.common import convert_robj


@pytest.fixture
def ObjectMixin(monkeypatch):
    from palladium.R import ObjectMixin
    r_dict = {}
    r = MagicMock()
    r.__getitem__.side_effect = r_dict.__getitem__
    r.__setitem__.side_effect = r_dict.__setitem__
    r['myfunc'] = Mock()
    r['predict'] = Mock()
    monkeypatch.setattr(ObjectMixin, 'r', r)
    return ObjectMixin


class TestDatasetLoader:
    @pytest.fixture
    def DatasetLoader(self, ObjectMixin):
        from palladium.R import DatasetLoader
        return DatasetLoader

    def test_it(self, DatasetLoader):
        X, y = object(), object()
        DatasetLoader.r['myfunc'].return_value = X, y
        dloader = DatasetLoader('myscript', 'myfunc', some='kwarg')
        assert dloader() == (X, y)
        dloader.r.source.assert_called_with('myscript')
        dloader.r['myfunc'].assert_called_with(some='kwarg')


class TestAbstractModel:
    @pytest.fixture
    def data(self):
        X = numpy.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
        y = numpy.array([1, 2])
        return X, y

    @pytest.fixture
    def dataframe(self, data):
        X, y = data
        return DataFrame(X, columns=['one', 'two', 'three']), Series(y)

    @pytest.fixture
    def Model(self, ObjectMixin, monkeypatch):
        from palladium.R import AbstractModel
        monkeypatch.setattr(AbstractModel, '__abstractmethods__', set())
        return AbstractModel

    def test_fit_with_numpy_data(self, Model, data):
        X, y = data
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.fit(X, y)
        funcargs = model.r['myfunc'].call_args
        assert (numpy.asarray(funcargs[0][0]) == X).all()
        assert (numpy.asarray(funcargs[0][1]) == y).all()
        assert funcargs[1]['some'] == 'kwarg'

    def test_fit_with_pandas_data(self, Model, dataframe):
        X, y = dataframe
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.fit(X, y)
        funcargs = model.r['myfunc'].call_args
        assert (convert_robj(funcargs[0][0]) == X).all().all()
        assert (convert_robj(funcargs[0][1]) == y).all()
        assert funcargs[1]['some'] == 'kwarg'


class TestClassificationModel(TestAbstractModel):
    @pytest.fixture
    def Model(self, ObjectMixin, monkeypatch):
        from palladium.R import ClassificationModel
        return ClassificationModel

    def test_predict_with_numpy_data(self, Model, data):
        X, y = data
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.r['predict'].return_value = numpy.array(
            [[0.1, 0.2, 0.7], [0.8, 0.1, 0.1]])
        model.fit(X, y)

        result = model.predict(X)
        predictargs = model.r['predict'].call_args
        assert predictargs[0][0] is model.rmodel_
        assert (numpy.asarray(predictargs[0][1]) == X).all()
        assert predictargs[1]['type'] == 'prob'
        assert (result ==
                numpy.argmax(model.r['predict'].return_value, axis=1)).all()

        result = model.predict_proba(X)
        assert (result == model.r['predict'].return_value).all()

    def test_predict_with_pandas_data(self, Model, dataframe):
        X, y = dataframe
        model = Model(scriptname='myscript', funcname='myfunc', some='kwarg')
        model.r['predict'].return_value = numpy.array(
            [[0.1, 0.2, 0.7], [0.8, 0.1, 0.1]])
        model.fit(X, y)

        result = model.predict(X)
        predictargs = model.r['predict'].call_args
        assert predictargs[0][0] is model.rmodel_
        assert (convert_robj(predictargs[0][1]) == X).all().all()
        assert predictargs[1]['type'] == 'prob'
        assert (result ==
                numpy.argmax(model.r['predict'].return_value, axis=1)).all()

        result = model.predict_proba(X)
        assert (result == model.r['predict'].return_value).all()


class TestClassification:
    @pytest.fixture
    def dataset(self):
        from palladium.R import DatasetLoader
        return DatasetLoader(
            scriptname=os.path.join(os.path.dirname(__file__), 'test_R.R'),
            funcname='dataset',
            )

    @pytest.fixture
    def model(self):
        from palladium.R import ClassificationModel
        return ClassificationModel(
            scriptname=os.path.join(os.path.dirname(__file__), 'test_R.R'),
            funcname='train.randomForest',
            encode_labels=True,
            )

    def test_fit_and_predict(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        probas = model.predict_proba(X)
        assert probas.shape == (150, 3)
        assert (probas.sum(axis=1) == 1).all()
        assert (model.predict(X) == numpy.asarray(y)).all()

    def test_fit_and_score(self, dataset, model):
        X, y = dataset()
        model.fit(X, y)
        assert model.score(X, y) == 1.0


class TestClassificationWithNumpyDataset(TestClassification):
    @pytest.fixture
    def dataset(self):
        X = numpy.array([
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
            ] * 50)
        y = numpy.array([1, 2, 3] * 50)
        return lambda: (X, y)


class TestClassificationWithPandasDataset(TestClassification):
    @pytest.fixture
    def dataset(self):
        X = DataFrame([
            [1.0, 2.0, 3.0],
            [4.0, 5.0, 6.0],
            [7.0, 8.0, 9.0],
            ] * 50)
        y = Series([1, 2, 3] * 50)
        return lambda: (X, y)



"""Utilities for testing the performance of a trained model.
"""

from pprint import pprint
import sys

from docopt import docopt

from .util import args_from_config
from .util import initialize_config
from .util import logger
from .util import timer


@args_from_config
def test(dataset_loader_test, model_persister, model_version=None):

    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_test()

    with timer(logger.info, "Reading model"):
        model = model_persister.read(version=model_version)

    logger.info(
        'Loaded model version {}'.format(model.__metadata__['version']))

    if not hasattr(model, 'score'):
        raise ValueError(
            "Your model doesn't seem to implement a 'score' method."
            )

    with timer(logger.info, "Applying model"):
        score = model.score(X, y)

    logger.info("Score: {}.".format(score))


def test_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
Test a model.

Usage:
  palladium-test [options]

Options:
  -h --help                  Show this screen.

  --model-version=<version>  The version of the model to be tested. If
                             not specified, the newest model will be used.
"""
    arguments = docopt(__doc__, argv=argv)
    model_version = arguments['--model-version']
    model_version = int(model_version) if model_version is not None else None
    initialize_config(__mode__='fit')
    test(model_version=model_version)


@args_from_config
def list(model_persister):
    pprint(model_persister.list())


def list_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
List information about available models.

Usage:
  palladium-list [options]

Options:
  -h --help                  Show this screen.
"""
    docopt(__doc__, argv=argv)
    initialize_config(__mode__='fit')
    list()



"""Utilities for fitting modles.
"""

import sys

from datetime import datetime
from docopt import docopt
from pprint import pformat
from sklearn.grid_search import GridSearchCV

from .interfaces import annotate
from .util import apply_kwargs
from .util import args_from_config
from .util import initialize_config
from .util import logger
from .util import PluggableDecorator
from .util import timer


@PluggableDecorator('fit_decorators')
@args_from_config
def fit(dataset_loader_train, model, model_persister, persist=True,
        dataset_loader_test=None, evaluate=False, persist_if_better_than=None):

    if persist_if_better_than is not None:
        evaluate = True
        if dataset_loader_test is None:
            raise ValueError(
                "When using 'persist_if_better_than', make sure you also "
                "provide a 'dataset_loader_test'."
                )

    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_train()

    with timer(logger.info, "Fitting model"):
        model.fit(X, y)

    if evaluate:
        with timer(logger.debug, "Evaluating model on train set"):
            score_train = model.score(X, y)
            annotate(model, {'score_train': score_train})
            logger.info("Train score: {}".format(score_train))

    score_test = None
    if evaluate and dataset_loader_test is not None:
        X_test, y_test = dataset_loader_test()
        with timer(logger.debug, "Evaluating model on test set"):
            score_test = model.score(X_test, y_test)
            annotate(model, {'score_test': score_test})
            logger.info("Test score:  {}".format(score_test))

    if persist:
        if (persist_if_better_than is not None and
            score_test < persist_if_better_than):
            logger.info("Not persisting model that has a test score "
                        "{} < {}".format(score_test, persist_if_better_than))
        else:
            annotate(model, {'train_timestamp': datetime.now().isoformat()})
            with timer(logger.info, "Writing model"):
                version = model_persister.write(model)
            logger.info("Wrote model with version {}.".format(version))

    return model


def fit_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
Fit a model and save to database.

Usage:
  palladium-fit [options]

Options:
  -n --no-save              Don't persist the fitted model to disk.

  --save-if-better-than=<k> Persist only if test score better than given
                            value.

  -e --evaluate             Evaluate fitted model on train and test set and
                            print out results.

  -h --help                 Show this screen.
"""
    arguments = docopt(__doc__, argv=argv)
    no_save = arguments['--no-save']
    save_if_better_than = arguments['--save-if-better-than']
    evaluate = arguments['--evaluate'] or bool(save_if_better_than)
    if save_if_better_than is not None:
        save_if_better_than = float(save_if_better_than)
    initialize_config(__mode__='fit')
    fit(
        persist=not no_save,
        evaluate=evaluate,
        persist_if_better_than=save_if_better_than,
        )


@args_from_config
def grid_search(dataset_loader_train, model, grid_search):
    with timer(logger.info, "Loading data"):
        X, y = dataset_loader_train()

    grid_search_kwargs = {
        'refit': False,
        }
    grid_search_kwargs.update(grid_search)

    cv = grid_search_kwargs.get('cv', None)
    if callable(cv):
        grid_search_kwargs['cv'] = apply_kwargs(cv, n=len(y), y=y)

    if not (hasattr(model, 'score') or 'scoring' in grid_search_kwargs):
        raise ValueError(
            "Your model doesn't seem to implement a 'score' method.  You may "
            "want to pass a 'scoring' argument to 'grid_search' instead."
            )

    with timer(logger.info, "Running grid search"):
        gs = GridSearchCV(model, **grid_search_kwargs)
        gs.fit(X, y)

    scores = sorted(gs.grid_scores_, key=lambda x: -x.mean_validation_score)
    logger.info("\n{}".format(pformat(scores)))
    return scores


def grid_search_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
Grid search parameters for the model.

Usage:
  palladium-grid-search [options]

Options:
  -h --help                Show this screen.
"""
    docopt(__doc__, argv=argv)
    initialize_config(__mode__='fit')
    grid_search()



from datetime import datetime
from functools import partial
import numpy as np
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch

from dateutil.parser import parse
import pytest


class TestFit:
    @pytest.fixture
    def fit(self):
        from palladium.fit import fit
        return fit

    @pytest.fixture
    def dataset_loader(self):
        dataset_loader = Mock()
        dataset_loader.return_value = Mock(), Mock()
        return dataset_loader

    def test_it(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        model_persister.write.assert_called_with(model)

    def test_no_persist(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister,
                     persist=False)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model_persister.call_count == 0

    def test_evaluate_no_test_dataset(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        X, y = object(), object()
        dataset_loader_train.return_value = X, y

        result = fit(dataset_loader_train, model, model_persister,
                     evaluate=True)

        assert result is model
        dataset_loader_train.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model.score.call_count == 1
        model.score.assert_called_with(X, y)
        model_persister.write.assert_called_with(model)

    def test_evaluate_with_test_dataset(self, fit):
        model, dataset_loader_train, model_persister = Mock(), Mock(), Mock()
        dataset_loader_test = Mock()
        X, y, X_test, y_test = object(), object(), object(), object()
        dataset_loader_train.return_value = X, y
        dataset_loader_test.return_value = X_test, y_test

        result = fit(dataset_loader_train, model, model_persister,
                     dataset_loader_test=dataset_loader_test,
                     evaluate=True)

        assert result is model
        dataset_loader_train.assert_called_with()
        dataset_loader_test.assert_called_with()
        model.fit.assert_called_with(X, y)
        assert model.score.call_count == 2
        assert model.score.mock_calls[0] == call(X, y)
        assert model.score.mock_calls[1] == call(X_test, y_test)
        model_persister.write.assert_called_with(model)

    def test_evaluate_annotations(self, fit, dataset_loader):
        model = Mock()
        model.score.side_effect = [0.9, 0.8]

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=Mock(),
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result.__metadata__['score_train'] == 0.9
        assert result.__metadata__['score_test'] == 0.8

    def test_persist_if_better_than(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result is model
        assert model_persister.write.call_count == 1

    def test_persist_if_better_than_false(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.91,
            )

        assert result is model
        assert model_persister.write.call_count == 0

    def test_persist_if_better_than_persist_false(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9

        result = fit(
            dataset_loader_train=dataset_loader,
            model=model,
            model_persister=model_persister,
            persist=False,
            dataset_loader_test=dataset_loader,
            persist_if_better_than=0.9,
            )

        assert result is model
        assert model_persister.write.call_count == 0

    def test_persist_if_better_than_no_dataset_test(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()
        model.score.return_value = 0.9

        with pytest.raises(ValueError):
            fit(
                dataset_loader_train=dataset_loader,
                model=model,
                model_persister=model_persister,
                dataset_loader_test=None,
                persist_if_better_than=0.9,
                )

    def test_timestamp(self, fit, dataset_loader):
        model, model_persister = Mock(), Mock()

        def persist(model):
            assert 'train_timestamp' in model.__metadata__
        model_persister.write.side_effect = persist

        before_fit = datetime.now()
        result = fit(
            dataset_loader,
            model,
            model_persister,
            )
        after_fit = datetime.now()

        assert result is model

        timestamp = parse(model.__metadata__['train_timestamp'])
        assert before_fit < timestamp < after_fit
        model_persister.write.assert_called_with(model)


class TestGridSearch:
    @pytest.fixture
    def grid_search(self):
        from palladium.fit import grid_search
        return grid_search

    def test_it(self, grid_search):
        model, dataset_loader_train = Mock(), Mock()
        grid_search_params = {'verbose': 4}
        X, y = object(), object()
        dataset_loader_train.return_value = X, y
        scores = [
            Mock(mean_validation_score=0.1),
            Mock(mean_validation_score=0.2),
            ]

        with patch('palladium.fit.GridSearchCV') as GridSearchCV:
            GridSearchCV().grid_scores_ = scores
            result = grid_search(
                dataset_loader_train, model, grid_search_params)

        assert result == list(reversed(scores))
        dataset_loader_train.assert_called_with()
        GridSearchCV.assert_called_with(model, refit=False, verbose=4)
        GridSearchCV().fit.assert_called_with(X, y)

    def test_no_score_method_raises(self, grid_search):
        model, dataset_loader_train = Mock(spec=['fit', 'predict']), Mock()
        dataset_loader_train.return_value = object(), object()

        with pytest.raises(ValueError):
            grid_search(dataset_loader_train, model, {})

    def test_grid_search(self, grid_search):
        model, dataset_loader_train = Mock(), Mock()
        dataset_loader_train.return_value = (
            np.random.random((10, 10)), np.random.random(10))

        CVIterator = Mock()

        def cv_iterator(n, p):
            return CVIterator(n=n, p=p)

        grid_search_params = {'cv': partial(cv_iterator, p=2)}

        scores = [
            Mock(mean_validation_score=0.1),
            Mock(mean_validation_score=0.2),
            ]
        with patch('palladium.fit.GridSearchCV') as GridSearchCV:
            GridSearchCV().grid_scores_ = scores
            grid_search(dataset_loader_train, model, grid_search_params)

        GridSearchCV.assert_called_with(model, refit=False,
                                        cv=CVIterator.return_value)
        CVIterator.assert_called_with(n=10, p=2)



"""Support for building models using the R programming language.
"""

from palladium.interfaces import DatasetLoader
from palladium.interfaces import Model
import numpy as np
from pandas import DataFrame
from pandas import Series
from pandas.rpy.common import convert_to_r_dataframe
import rpy2.robjects
from rpy2.robjects.numpy2ri import numpy2ri
from sklearn.metrics import accuracy_score
from sklearn.preprocessing import LabelEncoder


class ObjectMixin:
    r = rpy2.robjects.r

    def __init__(self, scriptname, funcname, **kwargs):
        self.scriptname = scriptname
        self.funcname = funcname
        self.r.source(scriptname)
        self.rfunc = self.r[funcname]
        self.kwargs = kwargs


class DatasetLoader(DatasetLoader, ObjectMixin):
    """A :class:`~palladium.interfaces.DatasetLoader` that calls an R
    function to load the data.
    """
    def __call__(self):
        X, y = self.rfunc(**self.kwargs)
        return X, y


class AbstractModel(Model, ObjectMixin):
    def __init__(self, encode_labels=False, *args, **kwargs):
        super(Model, self).__init__(*args, **kwargs)
        self.encode_labels = encode_labels

    @staticmethod
    def _from_python(obj):
        if isinstance(obj, DataFrame):
            obj = convert_to_r_dataframe(obj)
        elif isinstance(obj, Series):
            obj = numpy2ri(obj.values)
        elif isinstance(obj, np.ndarray):
            obj = numpy2ri(obj)
        return obj

    def fit(self, X, y=None):
        if self.encode_labels:
            self.enc_ = LabelEncoder()
            y = self.enc_.fit_transform(y)

        self.rmodel_ = self.rfunc(
            self._from_python(X),
            self._from_python(y),
            **self.kwargs)


class ClassificationModel(AbstractModel):
    """A :class:`~palladium.interfaces.Model` for classification problems
    that uses an R model for training and prediction.
    """

    def predict_proba(self, X):
        X = self._from_python(X)
        return np.asarray(self.r['predict'](self.rmodel_, X, type='prob'))

    def predict(self, X):
        X = X.astype(float) if hasattr(X, 'astype') else X
        y_pred = np.argmax(self.predict_proba(X), axis=1)
        if self.encode_labels:
            y_pred = self.enc_.inverse_transform(y_pred)
        return y_pred

    def score(self, X, y):
        return accuracy_score(self.predict(X), np.asarray(y))



from contextlib import contextmanager
import os
from unittest.mock import patch

import pytest


@pytest.fixture(autouse=True)
def config(request):
    from palladium.util import _config

    orig = _config.copy()
    _config.clear()
    _config.initialized = False
    request.addfinalizer(
        lambda: (_config.clear(), _config.update(orig)))
    return _config


@pytest.fixture
def flask_app(config):
    from palladium.server import app
    return app


@contextmanager
def change_cwd(path):
    cwd = os.getcwd()
    if path:
        os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


def predict(path='/predict?sepal length=1.0&sepal width=1.1&'
            'petal length=0.777&petal width=5'):
    from palladium.server import app
    from palladium.server import predict

    with app.test_request_context(path):
        response = predict()
        assert response.status_code == 200


def run_smoke_tests_with_config(config_fname, run=None, raise_errors=True,
                                func_kwargs=None):
    from palladium.util import logger  # don't break coverage
    from palladium.util import timer

    if func_kwargs is None:
        func_kwargs = {}

    print("Running functional tests for {}".format(config_fname))

    with patch.dict('os.environ', {'PALLADIUM_CONFIG': config_fname}):
        from palladium.fit import fit
        from palladium.fit import grid_search
        from palladium.eval import test
        from palladium.util import initialize_config

        failed = []

        with change_cwd(os.path.dirname(config_fname)):
            initialize_config(__mode__='fit')
            for func in (fit, grid_search, test, predict):
                if run and func.__name__ not in run:
                    continue
                try:
                    this_kwargs = func_kwargs.get(func.__name__, {})
                    with timer(logger.info,
                               "Running {}".format(func.__name__)):
                        func(**this_kwargs)
                except Exception as e:
                    if raise_errors:
                        raise
                    failed.append((func.__name__, e))

    return failed


def pytest_configure(config):
    # Don't allow accidental PALLADIUM_CONFIGs to leak into tests:
    os.environ.pop('PALLADIUM_CONFIG', None)


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", help="run slow tests")


def pytest_runtest_setup(item):
    if 'slow' in item.keywords and not item.config.getoption("--runslow"):
        pytest.skip("need --runslow option to run")



"""Assorted utilties.
"""

from collections import UserDict
from contextlib import contextmanager
from datetime import datetime
from functools import partial
from functools import wraps
import logging
from logging.config import dictConfig
from importlib import import_module
from inspect import getfullargspec
from inspect import getcallargs
import os
import sys
from threading import Thread
from time import sleep
from time import time

import dateutil.parser
import dateutil.rrule
from docopt import docopt
import psutil

from . import __version__


logger = logging.getLogger('palladium')


PALLADIUM_CONFIG_ERROR = """
  Maybe you forgot to set the environment variable PALLADIUM_CONFIG
  to point to your Palladium configuration file?  If so, please
  refer to the manual for more details.
"""


def resolve_dotted_name(dotted_name):
    if ':' in dotted_name:
        module, name = dotted_name.split(':')
    else:
        module, name = dotted_name.rsplit('.', 1)

    attr = import_module(module)
    for name in name.split('.'):
        attr = getattr(attr, name)

    return attr


def create_component(specification):
    specification = specification.copy()
    factory_dotted_name = specification.pop('__factory__')
    factory = resolve_dotted_name(factory_dotted_name)
    return factory(**specification)


class Config(dict):
    """A dictionary that represents the app's configuration.

    Tries to send a more user friendly message in case of KeyError.
    """
    initialized = False

    def __getitem__(self, name):
        try:
            return super(Config, self).__getitem__(name)
        except KeyError:
            raise KeyError(
                "The required key '{}' was not found in your "
                "configuration. {}".format(name, PALLADIUM_CONFIG_ERROR))

_config = Config()


def get_config(**extra):
    if not _config.initialized:
        _config.update(extra)
        _config.initialized = True
        fname = os.environ.get('PALLADIUM_CONFIG')
        if fname is not None:
            sys.path.insert(0, os.path.dirname(fname))

            with open(fname) as f:
                _config.update(
                    eval(f.read(), {'environ': os.environ})
                    )
                _initialize_config(_config)

    return _config


def initialize_config(**extra):
    if _config.initialized:
        raise RuntimeError("Configuration was already initialized")
    return get_config(**extra)


def _initialize_config_recursive(mapping):
    rv = []
    for key, value in tuple(mapping.items()):
        if isinstance(value, dict):
            rv.extend(_initialize_config_recursive(value))
            if '__factory__' in value:
                mapping[key] = create_component(value)
                rv.append(mapping[key])
        elif isinstance(value, (list, tuple)):
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    rv.extend(_initialize_config_recursive(item))
                    if '__factory__' in item:
                        value[i] = create_component(item)
                        rv.append(value[i])
    return rv


def _initialize_config(config):
    components = []

    if 'logging' in config:
        dictConfig(config['logging'])
    else:
        logging.basicConfig(level=logging.DEBUG)

    components = _initialize_config_recursive(config)
    for component in components:
        if hasattr(component, 'initialize_component'):
            component.initialize_component(config)

    return config


def apply_kwargs(func, **kwargs):
    """Call *func* with kwargs, but only those kwargs that it accepts.
    """
    new_kwargs = {}
    func_args = getfullargspec(func)[0]
    for i, argname in enumerate(func_args):
        if argname in kwargs:
            new_kwargs[argname] = kwargs[argname]
    return func(**new_kwargs)


def args_from_config(func):
    """Decorator that injects parameters from the configuration.
    """
    func_args = getfullargspec(func)[0]

    @wraps(func)
    def wrapper(*args, **kwargs):
        config = get_config()
        for i, argname in enumerate(func_args):
            if len(args) > i or argname in kwargs:
                continue
            elif argname in config:
                kwargs[argname] = config[argname]
        try:
            getcallargs(func, *args, **kwargs)
        except TypeError as exc:
            msg = "{}\n{}".format(exc.args[0], PALLADIUM_CONFIG_ERROR)
            exc.args = (msg,)
            raise exc
        return func(*args, **kwargs)

    wrapper.__wrapped__ = func
    return wrapper


@contextmanager
def timer(log=None, message=None):
    if log is not None:
        log("{}...".format(message))

    info = {}
    t0 = time()
    yield info

    info['elapsed'] = time() - t0
    if log is not None:
        log("{} done in {:.3f} sec.".format(message, info['elapsed']))


@contextmanager
def session_scope(session):
    """Provide a transactional scope around a series of operations."""
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


class ProcessStore(UserDict):
    def __init__(self, *args, **kwargs):
        self.mtime = {}
        super(ProcessStore, self).__init__(*args, **kwargs)

    def __setitem__(self, key, item):
        super(ProcessStore, self).__setitem__(key, item)
        self.mtime[key] = datetime.now()

    def __getitem__(self, key):
        return super(ProcessStore, self).__getitem__(key)

    def __delitem__(self, key):
        super(ProcessStore, self).__delitem__(key)
        del self.mtime[key]


process_store = ProcessStore()


class RruleThread(Thread):
    """Calls a given function in intervals defined by given recurrence
    rules (from `datetuil.rrule`).
    """
    def __init__(self, func, rrule, sleep_between_checks=60):
        """
        :param callable func:
          The function that I will call periodically.

        :param rrule rrule:

          The :class:`dateutil.rrule.rrule` recurrence rule that
          defines when I will do the calls.  See the `python-dateutil
          docs <https://labix.org/python-dateutil>`_ for details on
          how to define rrules.

          For convenience, I will also accept a dict instead of a
          `rrule` instance, in which case I will instantiate an rrule
          using the dict contents as keyword parameters.

        :param int sleep_between_checks:
          Number of seconds to sleep before I check again if I should
          run the function *func*.
        """
        super(RruleThread, self).__init__(daemon=True)
        if isinstance(rrule, dict):
            rrule = self._rrule_from_dict(rrule)
        self.func = func
        self.rrule = rrule
        self.sleep_between_checks = sleep_between_checks
        self.last_execution = datetime.now()
        self.alive = True

    @classmethod
    def _rrule_from_dict(cls, rrule):
        kwargs = rrule.copy()
        for key, value in rrule.items():
            # Allow constants in datetutil.rrule to be passed as strings
            if isinstance(value, str) and hasattr(dateutil.rrule, value):
                kwargs[key] = getattr(dateutil.rrule, value)

        dstart = kwargs.get('dtstart')
        if isinstance(dstart, str):
            kwargs['dtstart'] = dateutil.parser.parse(dstart)
        return dateutil.rrule.rrule(**kwargs)

    def run(self):
        while self.alive:
            now = datetime.now()
            if not self.rrule.between(self.last_execution, now):
                sleep(self.sleep_between_checks)
                continue

            self.last_execution = now

            try:
                self.func()
            except:
                logger.exception(
                    "Failed to call {}".format(self.func.__name__))


def memory_usage_psutil():
    """Return the current process memory usage in MB.
    """
    process = psutil.Process(os.getpid())
    mem = process.get_memory_info()[0] / float(2 ** 20)
    return mem


def version_cmd(argv=sys.argv[1:]):  # pragma: no cover
    __doc__ = """
Print the version number of Palladium.

Usage:
  palladium-version [options]

Options:
  -h --help                Show this screen.
"""
    docopt(__doc__, argv=argv)
    print(__version__)


class PluggableDecorator:
    def __init__(self, decorator_config_name):
        self.decorator_config_name = decorator_config_name
        self.wrapped = None

    def __call__(self, func):
        self.func = func

        def wrapper(*args, **kwargs):
            # The motivation here is that we want to defer loading the
            # configuration until the function is called for the first
            # time.
            if self.wrapped is None:
                func = self.func
                decorators = get_config().get(
                    self.decorator_config_name, [])
                self.decorators = [
                    resolve_dotted_name(dec) if isinstance(dec, str) else dec
                    for dec in decorators
                    ]
                orig_func = func
                for decorator in self.decorators:
                    func = decorator(func)
                if self.decorators:
                    self.wrapped = wraps(orig_func)(func)
                else:
                    self.wrapped = orig_func
            return self.wrapped(*args, **kwargs)

        return wraps(func)(wrapper)


def get_metadata(error_code=0, error_message=None, status='OK'):
    metadata = {
        'status': status,
        'error_code': error_code,
    }
    if error_message is not None:
        metadata['error_message'] = error_message
    metadata.update(get_config().get('service_metadata', {}))
    return metadata


def Partial(func, **kwargs):
    """Allows the use of partially applied functions in the
    configuration.
    """
    if isinstance(func, str):
        func = resolve_dotted_name(func)
    return partial(func, **kwargs)



import gzip
import json
import os
import pickle
from threading import Thread
from unittest.mock import Mock
from unittest.mock import MagicMock
from unittest.mock import patch
from time import sleep

import pytest


class Dummy:
    def __init__(self, **kwargs):
        vars(self).update(kwargs)

    def __eq__(self, other):
        return vars(self) == vars(other)


class TestFile:
    @pytest.fixture
    def File(self):
        from palladium.persistence import File
        return File

    def test_init_path_without_version(self, File):
        with pytest.raises(ValueError):
            filepersister = File('path_without')

    def test_read(self, File):
        with patch('palladium.persistence.File.list') as list,\
            patch('palladium.persistence.gzip.open') as open,\
            patch('palladium.persistence.pickle.load') as load:
            list.return_value = [{'version': 99}]
            open.return_value = MagicMock()
            result = File('/models/model-{version}').read()
            open.assert_called_with('/models/model-99.pkl.gz', 'rb')
            assert result == load.return_value
            load.assert_called_with(open.return_value.__enter__.return_value)

    def test_read_with_version(self, File):
        with patch('palladium.persistence.File.list') as list,\
            patch('palladium.persistence.gzip.open') as open,\
            patch('palladium.persistence.pickle.load') as load:
            list.return_value = [{'version': 99}]
            open.return_value = MagicMock()
            result = File('/models/model-{version}').read(432)
            open.assert_called_with('/models/model-432.pkl.gz', 'rb')
            assert result == load.return_value
            load.assert_called_with(open.return_value.__enter__.return_value)

    def test_read_no_model(self, File):
        with patch('palladium.persistence.File.list') as list:
            list.return_value = []
            f = File('/models/model-{version}')
            with pytest.raises(IOError):
                f.read()

    def test_read_no_model_with_given_version(self, File):
        with patch('palladium.persistence.File.list') as list:
            list.return_value = []
            f = File('/models/model-{version}')
            with pytest.raises(IOError):
                f.read(1)

    def test_write_no_model_files(self, File):
        with patch('palladium.persistence.File.list') as list,\
            patch('palladium.persistence.File._write_md') as write_md,\
            patch('palladium.persistence.gzip.open') as open,\
            patch('palladium.persistence.pickle.dump') as dump:
            list.return_value = []
            open.return_value = MagicMock()
            model = MagicMock()
            result = File('/models/model-{version}').write(model)
            open.assert_called_with('/models/model-1.pkl.gz', 'wb')
            dump.assert_called_with(
                model,
                open.return_value.__enter__.return_value,
                )
            write_md.assert_called_with([model.__metadata__])
            assert result == 1

    def test_write_with_model_files(self, File):
        with patch('palladium.persistence.File.list') as list,\
            patch('palladium.persistence.File._write_md') as write_md,\
            patch('palladium.persistence.gzip.open') as open,\
            patch('palladium.persistence.pickle.dump') as dump:
            list.return_value = [{'version': 99}]
            open.return_value = MagicMock()
            model = MagicMock()
            result = File('/models/model-{version}').write(model)
            open.assert_called_with('/models/model-100.pkl.gz', 'wb')
            dump.assert_called_with(
                model,
                open.return_value.__enter__.return_value,
                )
            write_md.assert_called_with([
                {'version': 99},
                model.__metadata__,
                ])
            assert result == 100

    def test_update_metadata(self, File):
        model = MagicMock(__metadata__={
            'existing': 'entry', 'version': 'overwritten'})

        with patch('palladium.persistence.File.list') as list,\
            patch('palladium.persistence.File._write_md') as write_md,\
            patch('palladium.persistence.gzip.open'),\
            patch('palladium.persistence.pickle.dump'):
            list.return_value = [{'version': 99}]
            File('/models/model-{version}').write(model)
            assert model.__metadata__ == {
                'existing': 'entry',
                'version': 100,
                }
            write_md.assert_called_with([
                {'version': 99},
                model.__metadata__,
                ])

    def test_list_no_metadata(self, File):
        with patch('palladium.persistence.os.path.exists') as exists:
            exists.return_value = False
            assert File('model-{version}').list() == []
            exists.assert_called_with('model-metadata.json')

    def test_list_with_metadata(self, File):
        with patch('palladium.persistence.os.path.exists') as exists,\
            patch('builtins.open') as open,\
            patch('palladium.persistence.json.load') as load:
            exists.return_value = True
            assert File('model-{version}').list() == load.return_value
            exists.assert_called_with('model-metadata.json')
            open.assert_called_with('model-metadata.json', 'r')

    def test_write_md(self, File):
        with patch('builtins.open') as open,\
            patch('palladium.persistence.json.dump') as dump:
            md = [{'hello': 'world'}]
            File('model-{version}')._write_md(md)
            open.assert_called_with('model-metadata.json', 'w')
            dump.assert_called_with(
                md,
                open.return_value.__enter__.return_value,
                indent=4,
                )


class TestDatabase:
    @pytest.fixture
    def Database(self):
        from palladium.persistence import Database
        return Database

    @pytest.fixture
    def database(self, request, Database):
        path = '/tmp/palladium.testing-{}.sqlite'.format(os.getpid())
        request.addfinalizer(lambda: os.remove(path))
        return Database('sqlite:///{}'.format(path), chunk_size=4)

    @pytest.fixture
    def dbmodel(self, database):
        from palladium.util import session_scope

        model = Dummy(
            name='mymodel',
            __metadata__={'some': 'metadata', 'version': 1},
            )

        model_blob = gzip.compress(pickle.dumps(model), compresslevel=0)
        chunk_size = 4
        chunks = [model_blob[i:i + chunk_size]
                  for i in range(0, len(model_blob), chunk_size)]

        dbmodel = database.DBModel(
            version=1,
            chunks=[
                database.DBModelChunk(
                    model_version=1,
                    blob=chunk,
                    )
                for chunk in chunks
                ],
            metadata_=json.dumps(model.__metadata__),
            )

        with session_scope(database.session) as session:
            session.add(dbmodel)

        return model

    def test_read_no_entry(self, database):
        with pytest.raises(IOError):
            database.read()

    def test_read_with_existing_entry(self, database, dbmodel):
        model = database.read()
        assert model == dbmodel

    def test_write_no_entry(self, database):
        model = Dummy(name='mymodel')
        database.write(model)
        assert database.read() == model
        assert model.__metadata__['version'] == 1

    def test_write_with_existing_entry(self, database, dbmodel):
        model = Dummy(name='mymodel')
        database.write(model)
        assert database.read() == model
        assert model.__metadata__['version'] == 2
        assert database.read(1) == dbmodel

    def test_concurrent_read_write(self, database, dbmodel):
        model = Dummy(name='mymodel')

        _read_success = True
        def read():
            nonlocal _read_success
            try:
                database.read()
            except:
                _read_success = False
                raise

        _write_success = True
        def write():
            nonlocal _write_success
            try:
                database.write(model)
            except:
                _write_success = False
                raise

        read_threads = [Thread(target=read) for i in range(5)]
        write_threads = [Thread(target=write) for i in range(5)]

        for th in write_threads:
            th.start()
        for th in read_threads:
            th.start()
        for th in write_threads + read_threads:
            th.join()

        assert _read_success
        assert _write_success

    def test_list_no_entries(self, database):
        assert database.list() == []

    def test_list_with_existing_entry(self, database, dbmodel):
        listing = database.list()
        assert listing == [dbmodel.__metadata__]

    def test_list_with_two_entries(self, database, dbmodel):
        model2 = Dummy(name='mymodel')
        database.write(model2)
        listing = database.list()
        assert listing == [dbmodel.__metadata__, model2.__metadata__]

class TestCachedUpdatePersister:
    @pytest.fixture
    def CachedUpdatePersister(self, process_store):
        from palladium.persistence import CachedUpdatePersister
        return CachedUpdatePersister

    @pytest.fixture
    def persister(self, CachedUpdatePersister, config):
        persister = CachedUpdatePersister(MagicMock())
        persister.initialize_component(config)
        return persister

    def test_read(self, process_store, persister):
        assert persister.read() is persister.impl.read.return_value

    def test_read_custom_value(self, process_store, persister):
        process_store['model'] = 'mymodel'
        assert persister.read() == 'mymodel'

    def test_write(self, persister):
        persister.write('mymodel')
        persister.impl.write.assert_called_with('mymodel')

    def test_update_cache(self, persister):
        persister.update_cache(version=123)
        assert persister.read() is persister.impl.read.return_value
        persister.impl.read.assert_called_with(version=123)

    def test_update_cache_rrule(self, process_store, CachedUpdatePersister,
                                config):
        rrule_info = {
            'freq': 'DAILY',
            'dtstart': '2014-10-30T13:21:18',
            }

        impl = MagicMock()
        persister = CachedUpdatePersister(impl, update_cache_rrule=rrule_info)
        persister.initialize_component(config)
        assert persister.read() is impl.read.return_value
        assert process_store['model'] is impl.read.return_value
        assert impl.read.call_count == 1

    def test_dont_cache(self, process_store, CachedUpdatePersister, config):
        config['__mode__'] = 'fit'

        rrule_info = {
            'freq': 'DAILY',
            'dtstart': '2014-10-30T13:21:18',
            }

        impl = MagicMock()
        persister = CachedUpdatePersister(impl, update_cache_rrule=rrule_info)
        persister.initialize_component(config)
        assert persister.read() is impl.read.return_value
        assert len(process_store) == 0
        assert persister.thread is None

    def test_list(self, CachedUpdatePersister):
        impl = Mock()
        persister = CachedUpdatePersister(impl)
        persister.list() is impl.load.return_value
